/*-
 * Copyright (c) 2017-2020 Carsten Sonne Larsen <cs@innolan.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"
#include "global.h"
#include "notify.h"
#include "message.h"
#include "setting.h"
#include "shutdown.h"
#include "sync.h"
#include "mem.h"
#include "win.h"
#include "tz.h"

#include "libraries/screennotify.h"

#include "logmod.h"
#define MODULENAME "Main"
#define EVT_HOTKEY 1L

static void MsgLoop(void);
static void CleanupBroker(void);

struct AppBroker
{
    CxObj *Object;
    struct MsgPort *BrokerPort;
    struct MsgPort *NotifyPort;
    struct MsgPort *ScreenPort;
    APTR screenNotifyWB;
    APTR screenNotifyPub;
};

static struct NewBroker newBroker = {
    NB_VERSION,
    (STRPTR)APP_SHORT_NAME,
    (STRPTR)APP_TITLE_VERSION,
    (STRPTR) "Synchronize Clock With Internet Servers",
    NBU_UNIQUE | NBU_NOTIFY,
    COF_SHOW_HIDE,
    0, 0, 0};

static struct AppBroker Broker;
static volatile bool BrokerRunning = false;

void StartBroker(void)
{
    LONG error;
    CxObj *filter, *sender, *translate;

    LogTrace("Initializing");
    Broker.Object = NULL;
    Broker.BrokerPort = NULL;
    Broker.NotifyPort = NULL;
    Broker.ScreenPort = NULL;
    Broker.screenNotifyWB = NULL;
    Broker.screenNotifyPub = NULL;
    Ports.BrokerPort = NULL;
    Ports.MemoryPort = NULL;

    Broker.BrokerPort = CreateMsgPort();
    if (Broker.BrokerPort == NULL)
    {
        LogError("Could not allocate broker port");
        return;
    }

    Ports.BrokerPort = CreateMsgPort();
    if (Ports.BrokerPort == NULL)
    {
        LogError("Could not allocate broker user port");
        CleanupBroker();
        return;
    }

    Broker.NotifyPort = CreateMsgPort();
    if (Broker.NotifyPort == NULL)
    {
        LogError("Could not allocate broker notify port");
        CleanupBroker();
        return;
    }

    Broker.ScreenPort = CreateMsgPort();
    if (Broker.ScreenPort == NULL)
    {
        LogError("Could not allocate broker screen port");
        CleanupBroker();
        return;
    }

    Ports.MemoryPort = CreateMsgPort();
    if (Ports.MemoryPort == NULL)
    {
        LogError("Could not allocate broker reply port");
        CleanupBroker();
        return;
    }

    newBroker.nb_Port = Broker.BrokerPort;
    newBroker.nb_Pri = Settings->Priority;

    Broker.Object = CxBroker(&newBroker, &error);
    if (!Broker.Object)
    {
        switch (error)
        {
        case CBERR_SYSERR:
            LogError("System problems (CBERR_SYSERR). Could not allocate broker object");
            break;
        case CBERR_DUP:
            LogWarn(APP_SHORT_NAME " is already running");
            break;
        default:
            LogError("Could not allocate broker object (error code: %ld)", error);
            break;
        }

        CleanupBroker();
        return;
    }

    filter = CxFilter(Settings->PopKey);
    if (filter == NULL)
    {
        LogError("Could not allocate broker filter object");
        CleanupBroker();
        return;
    }

    AttachCxObj(Broker.Object, filter);

    sender = CxSender(Broker.BrokerPort, EVT_HOTKEY);
    if (sender == NULL)
    {
        LogError("Could not allocate broker sender object");
        CleanupBroker();
        return;
    }

    AttachCxObj(filter, sender);

    translate = CxTranslate(NULL);
    if (translate == NULL)
    {
        LogError("Could not allocate broker translate object");
        CleanupBroker();
        return;
    }

    AttachCxObj(filter, translate);

    error = CxObjError(filter);
    if (error != 0)
    {
        switch (error)
        {
        case CBERR_SYSERR:
            LogWarn("Commodity filter problems (CBERR_SYSERR)");
            break;
        case COERR_BADFILTER:
            LogWarn("Commodity HOTKEY error");
            break;
        default:
            LogWarn("Commodity filter error (error code: %ld)", error);
            break;
        }
    }

    if (ScreenNotifyBase != NULL)
    {
        Broker.screenNotifyWB = AddWorkbenchClient(Broker.ScreenPort, 0);
        Broker.screenNotifyPub = AddPubScreenClient(Broker.ScreenPort, 0);
        LogTrace("AddWorkbenchClient");
    }

    ActivateNotifyPort(Broker.NotifyPort);
    StartSynchronizer();

    MsgLoop();

    CleanupBroker();
}

bool IsBrokerRunning(void)
{
    return BrokerRunning;
}
void SetBrokerPriority(int priority)
{
    SetCxObjPri(Broker.Object, priority);
}

static void Deactivate(void)
{
    LogDebug("ACTIVE changed: 1 -> 0");
    LogWarn("Disabling time synchronization");
    SendMessageWait(Ports.SyncerPort, ATK_DISABLE);
    ActivateCxObj(Broker.Object, 0);
}

static void Activate(void)
{
    LogDebug("ACTIVE changed: 0 -> 1");
    LogWarn("Enabling time synchronization");
    ActivateCxObj(Broker.Object, 1);
    StartSynchronizer();
}

static void ChangeTimezone(void)
{
    // TODO: Make thread safe
    CleanupTimezone();
    InitTimezone();
    SendMessageTo(Ports.WindowPort, Ports.MemoryPort, ATK_TZ_CHANGED);
}

static void ChangeTimezoneDelayedProc(void)
{
    // TODO: Make thread safe
    Delay(15 * 50);
    ReopenLocale();
    ChangeTimezone();
    SendMessageTo(Ports.WindowPort, Ports.MemoryPort, ATK_TZ_CHANGED);
}

static void ChangeTimezoneDelayed(void)
{
    CreateNewProcTags(
        NP_Entry, (IPTR)ChangeTimezoneDelayedProc,
        NP_StackSize, 16 * 1024,
        NP_Name, 0,
        NP_Input, 0,
        NP_Output, 0,
        NP_Error, 0,
        NP_CloseInput, FALSE,
        NP_CloseOutput, FALSE,
        NP_CloseError, FALSE,
        NP_WindowPtr, 0,
        NP_ConsoleTask, 0,
        NP_Cli, FALSE,
        TAG_DONE);
}

static bool HandleBrokerMessage(LONG msgid, ULONG msgtype)
{
    switch (msgtype)
    {
    case CXM_IEVENT:
        switch (msgid)
        {
        case EVT_HOTKEY:
            LogDebug("Show window");
            ShowSettingWindow();
            break;
        default:
            break;
        }
        break;
    case CXM_COMMAND:
        switch (msgid)
        {
        case CXCMD_DISABLE:
            Deactivate();
            break;
        case CXCMD_ENABLE:
            Activate();
            break;
        case CXCMD_KILL:
            LogInfo("Received shut down from commodity");
            return false;
            break;
        case CXCMD_UNIQUE:
            LogWarn("Commodity duplicate detected");
            LogDebug("Show window");
            ShowSettingWindow();
            break;
        case CXCMD_APPEAR:
            LogDebug("Show window");
            ShowSettingWindow();
            break;
        case CXCMD_DISAPPEAR:
            LogDebug("Hide window");
            SendMessageTo(Ports.WindowPort, Ports.MemoryPort, ATK_SHUTDOWN);
            break;
        default:
            break;
        }
    }

    return true;
}

static bool HandleUserMessage(long msg)
{
    switch (msg)
    {
    case ATK_STORE:
        SaveSettings(true);
        SaveSettings(false);
        SendMessageTo(Ports.WindowPort, Ports.MemoryPort, ATK_SHUTDOWN);
        break;
    case ATK_APPLY:
        SaveSettings(false);
        SendMessageTo(Ports.WindowPort, Ports.MemoryPort, ATK_SHUTDOWN);
        break;
    case ATK_UNDO:
        ApplyAppSettings(WindowSettings, true);
        SendMessageTo(Ports.WindowPort, Ports.MemoryPort, ATK_SHUTDOWN);
        break;
    case ATK_RESTART:
        SendMessageWait(Ports.SyncerPort, ATK_RESTART);
        break;
    case ATK_DISABLE:
        Activate();
        break;
    case ATK_ENABLE:
        Deactivate();
        break;
    case ATK_READONLY:
        LogDebug("READONLY changed: 1 -> 0");
        LogWarn("Entering read only mode. Clock is not adjusted in read-only mode");
        break;
    case ATK_READWRITE:
        LogDebug("READONLY changed: 0 -> 1");
        LogWarn("Leaving read only mode. Clock is adjusted again");
        break;
    case ATK_SHUTDOWN:
        return false;
        break;
    default:
        break;
    }

    return true;
}

static void MsgLoop(void)
{
    bool reopenWindow = false;
    bool loop = true;
    long shutdown = LOOP_RUNNING;
    ULONG brokerSigMask = (1 << Broker.BrokerPort->mp_SigBit);
    ULONG userSigMask = (1 << Ports.BrokerPort->mp_SigBit);
    ULONG replySigMask = (1 << Ports.MemoryPort->mp_SigBit);
    ULONG notifySigMask = (1 << Broker.NotifyPort->mp_SigBit);
    ULONG screenSigMask = (1 << Broker.ScreenPort->mp_SigBit);
    ULONG sigMask = brokerSigMask | notifySigMask | userSigMask |
                    screenSigMask | replySigMask | SIGBREAKF_CTRL_C;

    BrokerRunning = true;
    LogTrace("Loop started");

    ActivateCxObj(Broker.Object, 1);

    if (Settings->Popup)
    {
        ShowSettingWindow();
    }

    do
    {
        ULONG sigrcvd = Wait(sigMask);
        if (sigrcvd & brokerSigMask)
        {
            CxMsg *msg;
            while ((msg = (CxMsg *)GetMsg(Broker.BrokerPort)))
            {
                bool cont;
                LONG msgId = CxMsgID(msg);
                ULONG msgType = CxMsgType(msg);
                ReplyMsg((struct Message *)msg);

                cont = HandleBrokerMessage(msgId, msgType);
                if (!cont)
                {
                    shutdown = SHUTDOWN_REQUEST;
                }
            }
        }

        if (sigrcvd & userSigMask)
        {
            struct ApplicationMesage *msg;
            while ((msg = (struct ApplicationMesage *)GetMsg(Ports.BrokerPort)))
            {
                bool cont;
                long msgId = msg->MsgId;
                ReplyMsg((struct Message *)msg);

                cont = HandleUserMessage(msgId);
                if (!cont)
                {
                    loop = false;
                }
            }
        }

        if (sigrcvd & notifySigMask)
        {
            struct NotifyMessage *msg;
            while ((msg = (struct NotifyMessage *)GetMsg(Broker.NotifyPort)))
            {
                long msgType = msg->nm_NReq->nr_UserData;
                ReplyMsg((struct Message *)msg);
                switch (msgType)
                {
                case ATK_TZ_CHANGED:
                    LogDebug("TZ changed");
                    ChangeTimezone();
                    break;
                case ATK_TZONE_CHANGED:
                    LogDebug("TZONE changed");
                    ChangeTimezone();
                    break;
                case ATK_LOCALE_CHANGED:
                    LogError("Locale changed");
                    ChangeTimezoneDelayed();
                    break;
                }
                break;
            }
        }

        if (sigrcvd & screenSigMask)
        {
            struct ScreenNotifyMessage *msg;
            while ((msg = (struct ScreenNotifyMessage *)GetMsg(Broker.ScreenPort)))
            {
                LogDebug("ScreenNotifyMessage");
                if (msg->snm_Type == SCREENNOTIFY_TYPE_WORKBENCH)
                {
                    if (msg->snm_Value == NULL)
                    {
                        LogDebug("WB Close");
                        LogDebug("Hide window");
                        SendMessageTo(Ports.WindowPort,
                                      Ports.MemoryPort,
                                      ATK_DISABLE);
                        reopenWindow = WindowProcRunning;
                    }
                    else
                    {
                        LogDebug("WB Open");
                        if (reopenWindow)
                        {
                            ShowSettingWindow();
                        }
                        ReopenLocale();
                        ChangeTimezone();
                        reopenWindow = false;
                    }
                }
                ReplyMsg((struct Message *)msg);
            }
        }

        if (sigrcvd & replySigMask)
        {
            struct ApplicationMesage *msg;
            while ((msg = (struct ApplicationMesage *)GetMsg(Ports.MemoryPort)))
            {
                FreeMemSafe(msg);
            }
        }

        if (sigrcvd & SIGBREAKF_CTRL_C)
        {
            LogInfo("Received CTRL + C");
            shutdown = SHUTDOWN_REQUEST;
        }

        if (shutdown == SHUTDOWN_REQUEST)
        {
            struct ShutdownInfo si;
            si.BrokerPort = Ports.BrokerPort;
            si.SynchronizerPort = Ports.SyncerPort;
            si.SettingWindowPort = Ports.WindowPort;
            si.ShutdownMessage = ATK_SHUTDOWN;
            shutdown = SHUTTING_DOWN;
            StartShutdown(&si);
        }
    } while (loop);

    BrokerRunning = false;
    LogTrace("Loop exited");
}

static void CleanupBroker(void)
{
    struct ApplicationMesage *msg;

    if (Broker.screenNotifyWB != NULL)
    {
        LogTrace("ScreenNotify WB cleared");
        while (!RemWorkbenchClient(Broker.screenNotifyWB))
        {
            Delay(10);
        }
    }

    if (Broker.screenNotifyPub != NULL)
    {
        LogTrace("ScreenNotify PubScreen cleared");
        while (!RemPubScreenClient(Broker.screenNotifyPub))
        {
            Delay(10);
        }
    }

    CleanupNotifications();

    if (Broker.Object != NULL)
    {
        DeleteCxObjAll(Broker.Object);
        Broker.Object = NULL;
        LogTrace("CxObj cleared");
    }

    CleanupMsgPort(&Broker.BrokerPort);
    LogTrace("BrokerPort cleared");

    CleanupMsgPort(&Broker.NotifyPort);
    LogTrace("NotifyPort cleared");

    CleanupMsgPort(&Broker.ScreenPort);
    LogTrace("ScreenPort cleared");

    CleanupVolMsgPort(&Ports.BrokerPort);
    LogTrace("UserPort cleared");

    while ((msg = (struct ApplicationMesage *)GetMsg(Ports.MemoryPort)))
    {
        FreeMemSafe(msg);
    }
    LogTrace("ReplyPort Msg cleared");

    CleanupVolMsgPort(&Ports.MemoryPort);
    LogTrace("ReplyPort cleared");
}
