/*-
 * Copyright (c) 2017-2020 Carsten Sonne Larsen <cs@innolan.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <exec/ports.h>
#include <exec/nodes.h>
#include <proto/exec.h>
#include <stdarg.h>

#include "log.h"
#include "string.h"
#include "logger.h"

static void VLogLine(enum LogSeverity level, const char *module, const char *format, va_list ap)
{
    struct MsgPort *port, *replyPort;
    struct LogMessage message;

    replyPort = CreateMsgPort();
    if (replyPort == NULL)
    {
        return;
    }

    message.Msg.mn_Node.ln_Type = NT_MESSAGE;
    message.Msg.mn_Length = sizeof(struct LogMessage);
    message.Msg.mn_ReplyPort = replyPort;
    message.Type = LogMessageText;
    message.Severity = level;
    StrCopy(message.Module, module);
    VSNPrintf(message.Text, MAXLOGLINESIZE - 1, format, ap);
    VSNPrintf(message.Text, MAXLOGLINESIZE - 1, format, ap);

    Forbid();
    port = FindPort((CONST_STRPTR)LOGGER_PORT_NAME);
    if (port != NULL)
    {
        PutMsg(port, (struct Message *)&message);
    }
    Permit();

    if (port != NULL)
    {
        WaitPort(replyPort);
        GetMsg(replyPort);
    }

    DeleteMsgPort(replyPort);
}

void LogMemTrace(const char *module, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    VLogLine(MemTraceMessage, module, format, args);
    va_end(args);
}

void LogDebug(const char *module, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    VLogLine(DebugMessage, module, format, args);
    va_end(args);
}

void LogTrace(const char *module, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    VLogLine(TraceMessage, module, format, args);
    va_end(args);
}

void LogInfo(const char *module, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    VLogLine(InfoMessage, module, format, args);
    va_end(args);
}

void LogWarn(const char *module, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    VLogLine(WarningMessage, module, format, args);
    va_end(args);
}

void LogError(const char *module, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    VLogLine(ErrorMessage, module, format, args);
    va_end(args);
}
